#!/usr/bin/env python

# Christopher Potts
# For CS224u / Linguist 188/288, Stanford

import csv
from copy import copy
from operator import itemgetter
import numpy as np
from numpy import array, dot, diag
from numpy.linalg import svd

######################################################################
# SIMPLE MATRIX CLASS USING CSV INPUT

class Matrix:
    def __init__(self, filename):
        self.filename = filename
        self.mat = []        
        reader = csv.reader(file(filename))
        header = reader.next()
        self.colnames = header[1: ]
        self.rownames = []
        for line in reader:
            self.rownames.append(line[0])            
            self.mat.append(array(map(float, line[1: ])))
        self.mat = array(self.mat)

    def get_word_index(self, word):
        if word not in self.rownames:
            raise ValueError('%s is not represented in %s' % (word, self.filename))
        return self.rownames.index(word)

    def get_vector_by_word(self, word):
        index = self.get_word_index(word)
        return self.mat[index, ]

######################################################################
# DISTANCES

def vector_length(u):
    return np.sqrt(dot(u, u))
    
def cosine_distance(u, v):
    return 1.0 - (dot(u, v) / (vector_length(u) * vector_length(v)))

def euclidean_distance(u, v):
    diff = u - v
    return vector_length(diff)

######################################################################
# WEIGHTING

def length_norm(u):
    return u / vector_length(u)

def pmi(m, positive=False, discounting=False):    
    # Joint probability table:
    p = m.mat / np.sum(m.mat, axis=None)
    # PMI with positive option:
    colprobs = np.sum(p, axis=0)    
    np_pmi_log = np.vectorize((lambda x : pmi_log(x, positive=positive)))
    p = array([np_pmi_log(row / (np.sum(row)*colprobs)) for row in p])
    # Optional discounting:
    if discounting:
        colsums = np.sum(m.mat, axis=0)
        fmatrix = m.mat / (m.mat + 1)
        dmatrix = array([pmi_discounting_row_func(row, colsums) for row in m.mat])        
        p *= fmatrix * dmatrix
    reweighted = copy(m)
    reweighted.mat = p
    return reweighted    

def pmi_log(x, positive=False):
    if x <= 0.0:
        return 0.0
    else:
        x = np.log2(x)
        if positive and x < 0.0:
            x = 0.0
        return x

def pmi_discounting_row_func(row, colsums):
    mincontext = np.minimum(np.sum(row), colsums)
    return mincontext / (mincontext + 1)
            
def tfidf(m):
    colsums = np.sum(m.mat, axis=0)
    doccount = m.mat.shape[1]
    w = array([tfidf_row_func(row, colsums, doccount) for row in m.mat])
    reweighted = copy(m)
    reweighted.mat = w
    return reweighted

def tfidf_row_func(row, colsums, doccount):
    df = float(len([x for x in row if x > 0]))
    idf = 0.0
    if df > 0.0 and df != doccount:
        idf = np.log(doccount / df)
    tfs = row/colsums
    return tfs * idf

######################################################################
# DIMENSIONALITY REDUCTION
        
def lsa(m, k=100):
    rowmat, singvals, colmat = svd(m.mat, full_matrices=False)
    singvals = diag(singvals)
    trunc = dot(rowmat[:, 0:k], singvals[0:k, 0:k])
    dimreduced = copy(m)
    dimreduced.mat = trunc
    dimreduced.colnames = ['d%s' % i for i in range(k)]
    return dimreduced    

######################################################################
# EXPLORATION

def neighbors(m, word, distfunc=cosine_distance, n=5):
    u = m.get_vector_by_word(word)
    dists = {m.rownames[i]: distfunc(u, m.mat[i]) for i in xrange(m.mat.shape[0])}    
    dists = sorted(dists.items(), key=itemgetter(1), reverse=False)
    if n:
        return dists[ : n]
    else:
        return dists

def semantic_orientation(m,
                         seeds1=['bad', 'nasty', 'poor', 'negative', 'unfortunate', 'wrong', 'inferior'],
                         seeds2=['good', 'nice', 'excellent', 'positive', 'fortunate', 'correct', 'superior'],
                         distfunc=cosine_distance):
    sm1 = semantic_orientation_seed_matrix(m, seeds1)
    sm2 = semantic_orientation_seed_matrix(m, seeds2)
    scores = {m.rownames[i]: semantic_orientation_row_func(m.mat[i], sm1, sm2, distfunc) for i in xrange(m.mat.shape[0])}
    return sorted(scores.items(), key=itemgetter(1), reverse=False)

def semantic_orientation_seed_matrix(m, seeds):
    indices = []
    for w in seeds:
        try:
            indices.append(m.get_word_index(w))
        except:
            pass
    if not indices:
        raise ValueError('The file %s contains no members of the seed set: %s' % (m.filename, ",".join(seeds)))
    return m.mat[array(indices), : ]

def semantic_orientation_row_func(row, sm1, sm2, distfunc):
    val1 = np.sum([distfunc(row, srow) for srow in sm1])
    val2 = np.sum([distfunc(row, srow) for srow in sm2])
    return val1 - val2
